$TITLE ShortestPathStartPoint OA4202 2010
$INLINECOM { }
 OPTIONS
   SOLPRINT =     OFF,
   DECIMALS =       1,
   LIMCOL   =      10,
   LIMROW   =      10,
   RESLIM   =      60, {max seconds}
   ITERLIM  =99999999, {max iterations}
   LP       =     CPLEX,
   MIP      =     CPLEX
 ;


$offlisting

$ontext
============================================================================
With the following few lines of code, we are reading in the network specification.
- We read in the set of nodes, from the nodes.csv file
- We read in the set of all arcs, from the arcs_set.csv file
- We define the subset of arcs which we can interdict
- Finally, we read in the arc data, from the arcs_data.csv file
============================================================================
$offtext


SET
    n   nodes  /
$INCLUDE  sample_nodes.csv
 /;

alias(n,i,j)   {(i,j) is for arc distance and flow}

SET
   arcs(i,j)         arclist  /
$ONDELIM
$INCLUDE  sample_arcs_set.csv
$OFFDELIM
 /;

SET
   attackarcs(i,j)         the arcs on which we can place interdictions  /
$ONDELIM
$INCLUDE  sample_attack_arcs_set.csv
$OFFDELIM
 /;


table arcdata(i,j,*)
$ONDELIM
$INCLUDE  sample_arcs_data.csv
$OFFDELIM
;

$ontext
============================================================================
Here, we  compute C_max and n * C_max.  The index Distance into the arcdata
table was read in by GAMS form the header of the arcs_data.csv file.
============================================================================
$offtext

SCALARS nC ;
 nC=0.0 ;

 LOOP(arcs(i,j),
   IF( nC<ABS(arcdata(i,j,'Distance')),
       nC=ABS(arcdata(i,j,'Distance')) ;
   );
 );

nC = nC * card(n);

display nC;

$ontext
============================================================================
Lets define the primal shortest path LP.

Our definition includes the parameters xbar(i,j).  We'll use them later, to
re-solve for the shortest path, given an interdiction plan X.  For now, we'll
just set the xbar(i,j) to zero.

Finally, we'll also introduce a variable NOPATH, that makes the LP feasible
even if there is no path from the start node to the end node.  You can think of
NOPATH as adding an extra, super-expensive arc between Start and End.  We'll
make the cost of this arc nC, so it doesn't get used unless there is no other
path from Start to End.
============================================================================
$offtext

 PARAMETERS
   delay
   xbar(i,j)
 ;

 delay = nC;

 LOOP(arcs(i,j),
   xbar(i,j)=0 ;
 );

 VARIABLE
   Z                 objective function value
 ;

 POSITIVE VARIABLES
   Y(i,j)            flow on arc i j
   NOPATH            artificial arc from Start to End
 ;

 EQUATIONS
   OBJECTIVE
   NETFLOW(n)
 ;

 OBJECTIVE..
   Z =e= SUM(arcs(i,j),(  arcdata(i,j,'Distance') + delay*xbar(i,j)$(attackarcs(i,j) ) ) *Y(i,j) )
       + nC*NOPATH
 ;

 NETFLOW(j)..
   SUM(arcs(i,j),Y(i,j)) - SUM(arcs(j,i),Y(j,i))      {+ flow in - flow out}
     =e= +(NOPATH-1)       $diag(j,'Start')   {start node}
         +(1-NOPATH)       $diag(j,'End')   {end node}
 ;

 MODEL ShortestPath
 /
   OBJECTIVE
   NETFLOW
 /;

$ontext
============================================================================
Now, lets solve for the shortest path, when there are no interdictions.
============================================================================
$offtext

 SOLVE ShortestPath USING LP MINIMIZING Z;       { find the uninterdicted shortest path }

$ontext
============================================================================
Lets print the uninterdicted shortest path solution to a file called modelOutput.out
============================================================================
$offtext


 file out /modelOutput.out/;
 put out;

 put '****' / ;
 put 'solve ShortestPath with no interdictions' / ;
 LOOP(arcs(i,j)$(Y.l(i,j)>0),
   put '   transit arc ',i.tl,' -> ', j.tl / ;
 );
 put 'transit cost=',Z.l:10:1 / ;


$ontext
============================================================================
Now, lets define the MIP for computing interdictions.

We'll introduce a scalar, 'nAttacks', that specifies the number of
interdictions to compute.

We want to maximize the shortest path.  So, the problem we want to solve is a
'max min' problem.  To put this into the solver, we'll take the dual of the
inner problem, so the problem becomes a 'max max'.  That is why in the
following code, you'll see the dual of the shortest path problem.

The binary variables X(i,j), will specify whether we attack arc i j.
============================================================================
$offtext


SCALARS nAttacks;

nAttacks = 1;

 POSITIVE VARIABLES
   Pi(n)              dual variables giving distance to node n
 ;

 BINARY VARIABLES
   X(i,j)            should we attack arc i j
 ;

 EQUATIONS
   DUAL_OBJECTIVE
   DUAL_CONSTRAINTS(i,j)
   EXTRA_NOPATH_ARC_CONSTRAINT
   ROADBLOCK_LIMIT
 ;

 DUAL_OBJECTIVE..
   Z =e= Pi('End') - Pi('Start')   {end node minus start node}
 ;

 DUAL_CONSTRAINTS(arcs(i,j))..
   Pi(j)-Pi(i) =l= arcdata(i,j,'Distance')+delay*X(i,j)$(attackarcs(i,j))
 ;

 EXTRA_NOPATH_ARC_CONSTRAINT..
   Pi('End') - Pi('Start') =l= nC   {Comes from the extra arc we added}
 ;

 ROADBLOCK_LIMIT..
   SUM(attackarcs(i,j),X(i,j)) =l= nAttacks
 ;
 
 MODEL BlockPlacementMIP
 /
   DUAL_OBJECTIVE
   DUAL_CONSTRAINTS
   EXTRA_NOPATH_ARC_CONSTRAINT
   ROADBLOCK_LIMIT
 /;

$ontext
============================================================================
Now, lets solve for the best interdiction plan.
============================================================================
$offtext

 BlockPlacementMIP.optfile = 1;

 SOLVE BlockPlacementMIP USING MIP MAXIMIZING Z;       { Find the best interdiction plans }

$ontext
============================================================================
Lets print the interdiction plan into the output file we had opened, modelOutput.out
============================================================================
$offtext

display X.l;

 put '****' / ;
 put 'interdiction plan with ', nAttacks, ' nAttacks:' / ;
 LOOP(arcs(i,j)$(X.l(i,j)>0),
   put '   attack arc: ',i.tl,' -> ', j.tl / ;
 );
 put 'transit cost with interdictions in place =',Z.l:10:1 / ;

$ontext
============================================================================
Finally, now that we have computed attack locations, lets re-solve for 
the operator's new best route.
============================================================================
$offtext

 LOOP(arcs(i,j),                        { transfer the interdiction solution to constant parameter }
   xbar(i,j)=X.l(i,j) ;
 );

 SOLVE ShortestPath USING LP MINIMIZING Z;       { find shortest escape route the interdictions in place }

 put '****' / ;
 put 'solve ShortestPath with interdictions in place' / ;
 put 'operator best response:' / ;
 LOOP(arcs(i,j)$(Y.l(i,j)>0),
   put '   transit arc ',i.tl,' -> ', j.tl / ;
 );
 IF(NOPATH.l>0,
   put ' There is no path.' / ;
 );
 put 'transit cost=',Z.l:10:1 / ;
